import random
import string
from kbucket import client as kb
from pairio import client as pa
from .sf_sort_recording import sf_sort_recording
from .sf_summarize_recording import sf_summarize_recording

def clear_job_results(*,batch_name,incomplete_only=True):
  batch=kb.loadObject(key=dict(batch_name=batch_name))
  jobs=batch['jobs']
  for job in jobs:
    clear_job_result(job,incomplete_only=incomplete_only)

def download_recordings(*,batch_name):
  batch=kb.loadObject(key=dict(batch_name=batch_name))
  jobs=batch['jobs']
  for job in jobs:
    print('DOWNLOADING: '+job['label'])
    dsdir=job['recording']['directory']
    kb.realizeFile(dsdir+'/raw.mda')
    
def run_jobs(*,batch_name):
  batch=kb.loadObject(key=dict(batch_name=batch_name))
  jobs=batch['jobs']
  for job in jobs:
    run_job(job)
    
def assemble_job_results(*,batch_name):
  batch=kb.loadObject(key=dict(batch_name=batch_name))
  jobs=batch['jobs']
  job_results=[]
  for job in jobs:
    print('ASSEMBLING: '+job['label'])
    result=kb.loadObject(key=job)
    if not result:
      raise Exception('Unable to load object for job: '+job['label'])
    job_results.append(dict(
        job=job,
        result=result
    ))
  print('Saving results...')
  kb.saveObject(key=dict(name='job_results',batch_name=batch_name),object=dict(job_results=job_results))
  print('Done.')

def clear_job_result(job,*,incomplete_only=True):
  val=pa.get(key=job)
  if val:
    if (not incomplete_only) or (val.startswith('in-process')) or (val.startswith('error')):
      print('Clearing job: '+job['label'])
      pa.set(key=job,value=None)

def do_sort_recording(job):
  return sf_sort_recording(sorter=job['sorter'],recording=job['recording'])

def do_summarize_recording(job):
  return sf_summarize_recording(recording=job['recording'])

def do_run_job(job):
  if job['command']=='sort_recording':
    return do_sort_recording(job)
  elif job['command']=='summarize_recording':
    return do_summarize_recording(job)
  else:
    return dict(error='Invalid job command: '+job['command'])

def run_job(job):
  val=pa.get(key=job)
  if val:
    return
  code=''.join(random.choice(string.ascii_uppercase) for x in range(10))
  if not pa.set(key=job,value='in-process-'+code,overwrite=False):
    return
  print('Running job: '+job['label'])
  result=do_run_job(job)
  val=pa.get(key=job)
  if val!='in-process-'+code:
    return
  if 'error' in result:
    print('Error running job: '+result['error'])
    pa.set(key=job,value='error-'+code)
    kb.save(key=dict(job=job,name='error'),value=result)
    return
  kb.saveObject(key=job,object=result)




import numpy as np
import json
import mlprocessors as mlpr
import spikeextractors as si
import spikewidgets as sw
from kbucket import client as kb

def write_json_file(fname,obj):
  with open(fname, 'w') as f:
    json.dump(obj, f)

def get_random_spike_waveforms(*,recording,sorting,unit,snippet_len,max_num,channels=None):
    st=sorting.getUnitSpikeTrain(unit_id=unit)
    num_events=len(st)
    if num_events>max_num:
        event_indices=np.random.choice(range(num_events),size=max_num,replace=False)
    else:
        event_indices=range(num_events)

    spikes=recording.getSnippets(reference_frames=st[event_indices].astype(int),snippet_len=snippet_len,channel_ids=channels)
    if spikes:
      spikes=np.dstack(tuple(spikes))
    else:
      spikes=np.zeros((recording.getNumChannels(),snippet_len,0))
    return spikes

def compute_unit_templates(*,recording,sorting,unit_ids,snippet_len=50):
    ret=[]
    for unit in unit_ids:
        #print('Unit {} of {}'.format(unit,len(unit_ids)))
        waveforms=get_random_spike_waveforms(recording=recording,sorting=sorting,unit=unit,snippet_len=snippet_len,max_num=100)
        template=np.mean(waveforms,axis=2)
        ret.append(template)
    return ret

def compute_template_snr(template,channel_noise_levels):
    channel_snrs=[]
    for ch in range(template.shape[0]):
        channel_snrs.append((np.max(template[ch,:])-np.min(template[ch,:]))/channel_noise_levels[ch])
    return float(np.max(channel_snrs))
    
def compute_channel_noise_levels(recording):
    channel_ids=recording.getChannelIds()
    M=len(channel_ids)
    X=recording.getTraces(start_frame=0,end_frame=int(np.minimum(1000,recording.getNumFrames())))
    ret=[]
    for ii in range(len(channel_ids)):
        noise_level=np.std(X[ii,:])
        ret.append(noise_level)
    return ret

class MemoryRecordingExtractor(si.RecordingExtractor):
    def __init__(self, parent_recording):
        si.RecordingExtractor.__init__(self)
        self._parent_recording=parent_recording
        self._traces=parent_recording.getTraces()
        self.copyChannelProperties(parent_recording)
        self._channel_index_map=dict()
        ids=parent_recording.getChannelIds()
        for ii,id in enumerate(ids):
          self._channel_index_map[id]=ii
        self._copy_channel_properties(parent_recording)

    def getTraces(self, channel_ids=None, start_frame=None, end_frame=None):
        if start_frame is None:
            start_frame=0
        if end_frame is None:
            end_frame=self.getNumFrames()
        if channel_ids is None:
            channel_ids=self.getChannelIds()
        channel_indices=[]
        for id in channel_ids:
          channel_indices.append(self._channel_index_map[id])
        return self._traces[channel_indices,start_frame:end_frame]

    def getChannelIds(self):
        return self._parent_recording.getChannelIds()

    def getNumFrames(self):
        return self._parent_recording.getNumFrames()

    def getSamplingFrequency(self):
        return self._parent_recording.getSamplingFrequency()

    def frameToTime(self, frame):
        return self._parent_recording.frameToTime(frame)

    def timeToFrame(self, time):
        return self._parent_recording.timeToFrame(time)

    def _copy_channel_properties(self, recording, channel_ids=None):
        if channel_ids is None:
            channel_ids=recording.getChannelIds()
        for id in channel_ids:
          pnames=recording.getChannelPropertyNames(channel_id=id)
          for pname in pnames:
            val=recording.getChannelProperty(channel_id=id,property_name=pname)
            self.setChannelProperty(channel_id=id,property_name=pname,value=val)

class ComputeUnitsInfo(mlpr.Processor):
  NAME='ComputeUnitsInfo'
  VERSION='0.1.1'
  recording_dir=mlpr.Input(directory=True,description='Recording directory')
  channel_ids=mlpr.IntegerListParameter(description='List of channels to use.',optional=True,default=[])
  unit_ids=mlpr.IntegerListParameter(description='List of units to use.',optional=True,default=[])
  firings=mlpr.Input(description='Firings file')
  json_out=mlpr.Output('The info as a .json file')
  
  def run(self):
    R0=si.MdaRecordingExtractor(dataset_directory=self.recording_dir,download=True)
    if (self.channel_ids) and (len(self.channel_ids)>0):
      R0=si.SubRecordingExtractor(parent_recording=R0,channel_ids=self.channel_ids)
    recording=sw.lazyfilters.bandpass_filter(recording=R0,freq_min=300,freq_max=6000)
    sorting=si.MdaSortingExtractor(firings_file=self.firings)
    ef=int(1e6)
    recording_sub=si.SubRecordingExtractor(parent_recording=recording,start_frame=0,end_frame=ef)
    recording_sub=MemoryRecordingExtractor(parent_recording=recording_sub)
    sorting_sub=si.SubSortingExtractor(parent_sorting=sorting,start_frame=0,end_frame=ef)
    unit_ids=self.unit_ids
    if (not unit_ids) or (len(unit_ids)==0):
      unit_ids=sorting.getUnitIds()
  
    channel_noise_levels=compute_channel_noise_levels(recording=recording)
    print('computing templates...')
    templates=compute_unit_templates(recording=recording_sub,sorting=sorting_sub,unit_ids=unit_ids)
    print('.')
    ret=[]
    for i,unit_id in enumerate(unit_ids):
      template=templates[i]
      info0=dict()
      info0['unit_id']=int(unit_id)
      info0['snr']=compute_template_snr(template,channel_noise_levels)
      peak_channel_index=np.argmax(np.max(np.abs(template),axis=1))
      info0['peak_channel']=int(recording.getChannelIds()[peak_channel_index])
      train=sorting.getUnitSpikeTrain(unit_id=unit_id)
      info0['num_events']=int(len(train))
      info0['firing_rate']=float(len(train)/(recording.getNumFrames()/recording.getSamplingFrequency()))
      ret.append(info0)
    write_json_file(self.json_out,ret)
  
  
## return format can be 'json' or 'filename'
def compute_units_info(*,recording_dir,firings,channel_ids=[],unit_ids=[],return_format='json'):
    out=ComputeUnitsInfo.execute(recording_dir=recording_dir,firings=firings,unit_ids=unit_ids,channel_ids=channel_ids,json_out={'ext':'.json'}).outputs
    fname=out['json_out']
    if return_format=='filename':
      return fname
    else:
      fname=kb.realizeFile(fname)
      with open(fname) as f:
        return json.load(f)

def select_units_on_channels(recording_dir,firings,channels):
  info=compute_units_info(recording_dir=recording_dir,firings=firings)
  units=[]
  for info0 in info:
    if info0['peak_channel'] in channels:
      units.append(info0['unit_id'])
  return units


from .sf_batch import sf_batch_prepare, sf_batch_run, sf_batch_assemble
from .sf_summarize_recording import sf_summarize_recording
from .sf_sort_recording import sf_sort_recording
from .sf_batch2 import clear_job_results, download_recordings, run_jobs, assemble_job_results
from .compute_units_info import compute_units_info, select_units_on_channels



from kbucket import client as kb
from pairio import client as pa
import getpass
from copy import deepcopy
import json

def setKBucketConfig(*,config=None,collection=None,key=None,verbose=False,return_config=False):
  if key:
    if config:
      raise Exception('Cannot specify both key and config')
    config=pa.getObject(collection=collection,key=key)
  pa.setConfig(
    collections=config['pairio']['collections'],
    user=config['pairio']['user'],
    token=config['pairio']['token'],
    read_local=config['pairio']['read_local'],
    write_local=config['pairio']['write_local'],
    read_remote=config['pairio']['read_remote'],
    write_remote=config['pairio']['write_remote']
  )
  kb.setConfig(
    share_ids=config['kbucket']['share_ids'],
    upload_share_id=config['kbucket']['upload_share_id'],
    upload_token=config['kbucket']['upload_token'],
    load_local=config['kbucket']['load_local'],
    load_remote=config['kbucket']['load_remote'],
    save_remote=config['kbucket']['save_remote']
  )
  if verbose:
    config_display=deepcopy(config)
    if config_display['pairio']['token']:
      config_display['pairio']['token']='************'
    if config_display['kbucket']['upload_token']:
      config_display['kbucket']['upload_token']='************'
    print ('Set config:')
    print (json.dumps(config_display,indent=2))
  if return_config:
    return config

def kbucketConfigLocal(write=True):
  pa.setConfig(
      collections=[],
      user='',
      token='',
      read_local=True,write_local=False,read_remote=False,write_remote=False
  )
  kb.setConfig(
      share_ids=[],
      upload_share_id='',
      upload_token='',
      load_local=True,load_remote=False,save_remote=False
  )
  if write:
    pa.setConfig(
        write_local=True,
    )

def kbucketConfigRemote(*,name='spikeforest1-readonly',collection='spikeforest',password=None,ask_password=False,verbose=False,return_config=False):
  if ask_password:
    password=getpass.getpass('Enter password: ')
    if not password:
      raise Exception('No password entered')

  key=dict(name=name)
  if password:
    key2=dict(key=key,password=password)
  else:
    key2=key

  return setKBucketConfig(collection=collection,key=key2,verbose=verbose,return_config=return_config)

def kbucketConfigRemoteOld(*,user='spikeforest',share_id='spikeforest.spikeforest2',password=None,write=False):
  pa.setConfig(
      collections=[user],
      user='',
      token='',
      read_local=False,write_local=False,read_remote=True,write_remote=False
  )
  kb.setConfig(
      share_ids=[share_id],
      upload_share_id='',
      upload_token='',
      load_local=True,load_remote=True,save_remote=False
  )
  if write:
    if password is None:
      password=getpass.getpass('Enter the spikeforest password (or leave blank for read-only)')
    if not password:
      return
    pa.setConfig(
        user=user,
        token=pa.get(collection='spikeforest',key=dict(name='pairio_token',user=user,password=password)),
        write_remote=True
    )
    kb.setConfig(
        upload_share_id=share_id,
        upload_token=pa.get(collection='spikeforest',key=dict(name='kbucket_token',share_id=share_id,password=password)),
        save_remote=True
    )


from kbucket import client as kb
import spikeextractors as si
import spikeforest as sf
import spiketoolkit as st
import mlprocessors as mlpr
import os
import shutil
import random
import string

class MountainSort4(mlpr.Processor):
    NAME='MountainSort4'
    VERSION='4.0.1'
    
    recording_dir=mlpr.Input('Directory of recording',directory=True)
    channels=mlpr.IntegerListParameter(description='List of channels to use.',optional=True,default=[])
    firings_out=mlpr.Output('Output firings file')
    
    detect_sign=mlpr.IntegerParameter('Use -1, 0, or 1, depending on the sign of the spikes in the recording')
    adjacency_radius=mlpr.FloatParameter('Use -1 to include all channels in every neighborhood')
    freq_min=mlpr.FloatParameter(optional=True,default=300,description='Use 0 for no bandpass filtering')
    freq_max=mlpr.FloatParameter(optional=True,default=6000,description='Use 0 for no bandpass filtering')
    whiten=mlpr.BoolParameter(optional=True,default=True,description='Whether to do channel whitening as part of preprocessing')
    clip_size=mlpr.IntegerParameter(optional=True,default=50,description='')
    detect_threshold=mlpr.FloatParameter(optional=True,default=3,description='')
    detect_interval=mlpr.IntegerParameter(optional=True,default=10,description='Minimum number of timepoints between events detected on the same channel')
    noise_overlap_threshold=mlpr.FloatParameter(optional=True,default=0.15,description='Use None for no automated curation')
    
    def run(self):
        recording=si.MdaRecordingExtractor(self.recording_dir)
        if len(self.channels)>0:
          recording=si.SubRecordingExtractor(parent_recording=recording,channel_ids=self.channels)
        num_workers=os.environ.get('NUM_WORKERS',None)
        if num_workers:
            num_workers=int(num_workers)
        sorting=sf.sorters.mountainsort4(
            recording=recording,
            detect_sign=self.detect_sign,
            adjacency_radius=self.adjacency_radius,
            freq_min=self.freq_min,
            freq_max=self.freq_max,
            whiten=self.whiten,
            clip_size=self.clip_size,
            detect_threshold=self.detect_threshold,
            detect_interval=self.detect_interval,
            noise_overlap_threshold=self.noise_overlap_threshold,
            num_workers=num_workers
        )
        si.MdaSortingExtractor.writeSorting(sorting=sorting,save_path=self.firings_out)
    

class IronClust(mlpr.Processor):
    NAME='IronClust'
    VERSION='4.2.6'
    
    recording_dir=mlpr.Input('Directory of recording',directory=True)
    channels=mlpr.IntegerListParameter(description='List of channels to use.',optional=True,default=[])
    firings_out=mlpr.Output('Output firings file')
    
    detect_sign=mlpr.IntegerParameter('Use -1, 0, or 1, depending on the sign of the spikes in the recording')
    adjacency_radius=mlpr.FloatParameter('Use -1 to include all channels in every neighborhood')
    detect_threshold=mlpr.FloatParameter(optional=True,default=3,description='')
    prm_template_name=mlpr.StringParameter(optional=False,description='TODO')
    freq_min=mlpr.FloatParameter(optional=True,default=300,description='Use 0 for no bandpass filtering')
    freq_max=mlpr.FloatParameter(optional=True,default=6000,description='Use 0 for no bandpass filtering')
    merge_thresh=mlpr.FloatParameter(optional=True,default=0.98,description='TODO')
    pc_per_chan=mlpr.IntegerParameter(optional=True,default=3,description='TODO')
    
    def run(self):
        ironclust_src=os.environ.get('IRONCLUST_SRC',None)
        if not ironclust_src:
            raise Exception('Environment variable not set: IRONCLUST_SRC')
        code=''.join(random.choice(string.ascii_uppercase) for x in range(10))
        tmpdir=os.environ.get('TEMPDIR','/tmp')+'/ironclust-tmp-'+code
            
        try:
            recording=si.MdaRecordingExtractor(self.recording_dir)
            if len(self.channels)>0:
              recording=si.SubRecordingExtractor(parent_recording=recording,channel_ids=self.channels)
            if not os.path.exists(tmpdir):
                os.mkdir(tmpdir)
            sorting=sf.sorters.ironclust(
                recording=recording,
                tmpdir=tmpdir, ## TODO
                detect_sign=self.detect_sign,
                adjacency_radius=self.adjacency_radius,
                detect_threshold=self.detect_threshold,
                merge_thresh=self.merge_thresh,
                freq_min=self.freq_min,
                freq_max=self.freq_max,
                pc_per_chan=self.pc_per_chan,
                prm_template_name=self.prm_template_name,
                ironclust_src=ironclust_src
            )
            si.MdaSortingExtractor.writeSorting(sorting=sorting,save_path=self.firings_out)
        except:
            if os.path.exists(tmpdir):
                shutil.rmtree(tmpdir)
            raise
        shutil.rmtree(tmpdir)
        
class SpykingCircus(mlpr.Processor):
    NAME='SpykingCircus'
    VERSION='0.1.2'
    
    recording_dir=mlpr.Input('Directory of recording',directory=True)
    channels=mlpr.IntegerListParameter(description='List of channels to use.',optional=True,default=[])
    firings_out=mlpr.Output('Output firings file')
    
    detect_sign=mlpr.IntegerParameter(description='-1, 1, or 0')
    adjacency_radius=mlpr.FloatParameter(optional=True,default=100,description='Channel neighborhood adjacency radius corresponding to geom file')
    spike_thresh=mlpr.FloatParameter(optional=True,default=6,description='Threshold for detection')
    template_width_ms=mlpr.FloatParameter(optional=True,default=3,description='Spyking circus parameter')
    filter=mlpr.BoolParameter(optional=True,default=True)
    whitening_max_elts=mlpr.IntegerParameter(optional=True,default=1000,description='I believe it relates to subsampling and affects compute time')
    clustering_max_elts=mlpr.IntegerParameter(optional=True,default=10000,description='I believe it relates to subsampling and affects compute time')
    
    def run(self):
        code=''.join(random.choice(string.ascii_uppercase) for x in range(10))
        tmpdir=os.environ.get('TEMPDIR','/tmp')+'/ironclust-tmp-'+code
        
        num_workers=os.environ.get('NUM_WORKERS',2)
            
        try:
            recording=si.MdaRecordingExtractor(self.recording_dir)
            if len(self.channels)>0:
              recording=si.SubRecordingExtractor(parent_recording=recording,channel_ids=self.channels)
            if not os.path.exists(tmpdir):
                os.mkdir(tmpdir)
            sorting=sf.sorters.spyking_circus(
                recording=recording,
                output_folder=tmpdir,
                probe_file=None,
                file_name=None,
                detect_sign=self.detect_sign,
                adjacency_radius=self.adjacency_radius,
                spike_thresh=self.spike_thresh,
                template_width_ms=self.template_width_ms,
                filter=self.filter,
                merge_spikes=True,
                n_cores=num_workers,
                electrode_dimensions=None,
                whitening_max_elts=self.whitening_max_elts,
                clustering_max_elts=self.clustering_max_elts
            )
            si.MdaSortingExtractor.writeSorting(sorting=sorting,save_path=self.firings_out)
        except:
            if os.path.exists(tmpdir):
                shutil.rmtree(tmpdir)
            raise
        shutil.rmtree(tmpdir)

#sf.sorters.ironclust(*, recording, tmpdir, detect_sign=-1, adjacency_radius=-1, detect_threshold=5, merge_thresh=0.98, freq_min=300, freq_max=6000, pc_per_chan=3, prm_template_name, ironclust_src=None)
        
Processors=dict(
    MountainSort4=MountainSort4,
    IronClust=IronClust,
    SpykingCircus=SpykingCircus
)
        
def sf_sort_recording(sorter,recording):
    dsdir=recording['directory']
    sorting_params=sorter['params']
    processor_name=sorter['processor_name']
    if processor_name in Processors:
        SS=Processors[processor_name]
    else:
        raise Exception('No such sorter: '+processor_name)
        
    outputs=SS.execute(
        recording_dir=dsdir,
        channels=recording.get('channels',[]),
        firings_out=dict(ext='.mda'),
        **sorting_params
    ).outputs
    firings_out=kb.saveFile(outputs['firings_out'])
    firings_true_path=recording['directory']+'/firings_true.mda'
    if not kb.findFile(firings_true_path):
        firings_true_path=None
    result=dict(
        recording_name=recording['name'],
        study_name=recording['study'],
        sorter_name=sorter['name'],
        recording_dir=dsdir,
        channels=recording.get('channels',[]),
        units_true=recording.get('units_true',[]),
        firings_true=firings_true_path,
        sorting_params=sorting_params,
        sorting_processor_name=SS.NAME,
        sorting_processor_version=SS.VERSION,
        firings=firings_out
    )
    result['summary']=sf.summarizeSorting(result)
    if result.get('firings_true',None):
        result['comparison_with_truth']=sf.compareWithTruth(result)
    
    return result



#!/usr/bin/env python

import argparse
import spikeforest as sf
from kbucket import client as kb
import os
import json

def read_json_file(fname):
    with open(fname) as f:
        return json.load(f)

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description = 'Run SpikeForest batch processing')
    parser.add_argument('command',help='clear, prepare, run, assemble')
    parser.add_argument('batch_name',help='Name of the batch')
    args = parser.parse_args()

    batch_name=args.batch_name

    spikeforest_password=os.environ.get('SPIKEFOREST_PASSWORD','')
    if not spikeforest_password:
      raise Exception('Environment variable not set: SPIKEFOREST_PASSWORD')
    
    print('Loading batch: '+batch_name)
    sf.kbucketConfigRemote(share_id='spikeforest.spikeforest1',write=True,password=spikeforest_password)
    obj=kb.loadObject(key=dict(batch_name=batch_name))
    if not obj:
      raise Exception('Unable to find batches object.')

    command=args.command
    if command=='clear':
      sf.sf_batch.clear_job_results(batch_name=batch_name,incomplete_only=False)
    elif command=='prepare':
      sf.sf_batch.download_recordings(batch_name=batch_name)
      sf.sf_batch.clear_job_results(batch_name=batch_name,incomplete_only=True)
    elif command=='run':
      sf.sf_batch.run_jobs(batch_name=batch_name)
    elif command=='assemble':
      sf.sf_batch.assemble_job_results(batch_name=batch_name)
    else:
      raise Exception('Unrecognized command: '+command)


import numpy as np

def get_random_spike_waveforms(*,recording,sorting,unit,snippet_len,max_num,channels=None):
    st=sorting.getUnitSpikeTrain(unit_id=unit)
    num_events=len(st)
    if num_events>max_num:
        event_indices=np.random.choice(range(num_events),size=max_num,replace=False)
    else:
        event_indices=range(num_events)

    spikes=recording.getSnippets(reference_frames=st[event_indices].astype(int),snippet_len=snippet_len,channel_ids=channels)
    spikes=np.dstack(tuple(spikes))
    return spikes

def compute_unit_templates(*,recording,sorting,unit_ids,snippet_len=50):
    ret=[]
    for unit in unit_ids:
        waveforms=get_random_spike_waveforms(recording=recording,sorting=sorting,unit=unit,snippet_len=snippet_len,max_num=200)
        template=np.mean(waveforms,axis=2)
        ret.append(template)
    return ret

def compute_template_snr(template,channel_noise_levels):
    channel_snrs=[]
    for ch in range(template.shape[0]):
        channel_snrs.append((np.max(template[ch,:])-np.min(template[ch,:]))/channel_noise_levels[ch])
    return np.max(channel_snrs)
    
def compute_channel_noise_levels(recording):
    channel_ids=recording.getChannelIds()
    M=len(channel_ids)
    X=recording.getTraces(start_frame=0,end_frame=np.minimum(1000,recording.getNumFrames()))
    ret=[]
    for ii in range(len(channel_ids)):
        noise_level=np.std(X[ii,:])
        ret.append(noise_level)
    return ret

def compute_unit_snrs(*,recording,sorting,unit_ids=None):
    if unit_ids is None:
        unit_ids=sorting.getUnitIds()
    channel_noise_levels=compute_channel_noise_levels(recording=recording)
    templates=compute_unit_templates(recording=recording,sorting=sorting,unit_ids=unit_ids)
    ret=dict()
    for i,template in enumerate(templates):
        snr=compute_template_snr(template,channel_noise_levels)
        ret[unit_ids[i]]=snr
    return ret


import vdomr as vd

class SelectBox(vd.Component):
    def __init__(self,options=[]):
        vd.Component.__init__(self)
        self._on_change_handlers=[]
        self._value=None
        self.setOptions(options)
        
    def setOptions(self,options):
        self._options=options
        if self._value not in options:
          self._value=options[0] if options else None
        self.refresh()
        
    def value(self):
        return self._value
    
    def setValue(self,value):
        self._value=value
        self.refresh()
        
    def onChange(self,handler):
        self._on_change_handlers.append(handler)
        
    def _on_change(self,value):
        self._value=value
        for handler in self._on_change_handlers:
            handler(value=value)
        
    def render(self):
        opts=[]
        for option in self._options:
            if option==self._value:
              opts.append(vd.option(option,selected='selected'))
            else:
              opts.append(vd.option(option))
        X=vd.select(opts,onchange=self._on_change)
        return X

class SFSelectWidget(vd.Component):
  def __init__(self,sfdata,mode):
    vd.Component.__init__(self)
    self._SF=sfdata
    self._study_box=SelectBox()
    self._recording_box=SelectBox()
    self._sorting_result_box=SelectBox()
    self._mode=mode

    self._study_box.onChange(self._on_change)
    self._recording_box.onChange(self._on_change)
    self._sorting_result_box.onChange(self._on_change)

  def _on_change(self,value):
    self.refresh()
  def study(self):
    name=self._study_box.value()
    return self._SF.study(name) if name else None
  def recording(self):
    study=self.study()
    if not study:
      return None
    name=self._recording_box.value()
    return study.recording(name) if name else None
  def sortingResult(self):
    recording=self.recording()
    if not recording:
      return None
    name=self._sorting_result_box.value()
    return recording.sortingResult(name) if name else None
  def setStudyName(self,name):
    self._study_box.setValue(name)
  def setRecordingName(self,name):
    self._recording_box.setValue(name)
  def setSortingResultName(self,name):
    self._sorting_result_box.setValue(name)
  def render(self):
    self._study_box.setOptions(self._SF.studyNames())
    study=self.study()
    if study:
      self._recording_box.setOptions(study.recordingNames())
      recording=self.recording()
      if recording:
        self._sorting_result_box.setOptions(recording.sortingResultNames())
      else:
        self._sorting_result_box.setOptions([])
    else:
      self._recording_box.setOptions([])
      self._sorting_result_box.setOptions([])

    study_row=vd.tr(vd.td('Study **: '),vd.td(self._study_box))
    recording_row=vd.tr(vd.td('Recording: '),vd.td(self._recording_box))
    sorting_result_row=vd.tr(vd.td('Sorting result: '),vd.td(self._sorting_result_box))
    
    rows=[]
    mode=self._mode
    if mode=='study':
      rows=[study_row]
    elif mode=='recording':
      rows=[study_row,recording_row]
    elif mode=='sorting_result':
      rows=[study_row,recording_row,sorting_result_row]
    else:
      raise Exception('Invalid mode: '+mode)
    
    Table=vd.table(*rows)
    div=vd.div(
      vd.hr(),
      Table,
      vd.hr(),
    )

    return div